#define GLOBAL
#include <math.h>
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#include "utils.h"

#define DEFSITINI 34567	  /* Valeur du FlagExact pour modifier la rfrence  */

GLOBAL double OAi[3*9];	  /* Pour la visu par exemple */

/* On prend les constantes dans le common */
extern double Rpoul, ABi[], OHi[], OA_ref[];                
extern int flagbas[];                               /* 1 pour poulies basses */

/*---------------------------------------------------------------------------*/
/*                  Petit common  local                                      */
/*---------------------------------------------------------------------------*/
static double Qzero[9];
/* Flag  1 si les constantes du modle simplifi sont pas encore calcules  */
static int FlagNotiniConstSmp = 1;

/*---------------------------------------------------------------------------*/
/*            Modele geometrique exact et simplifie                          */
/*---------------------------------------------------------------------------*/
/*  Entrees :                                                                */
/*      OA[3*4]  : Position-Orientation de la maquette                      */
/*                                                                           */
/*      FlagExact : 0 pour modele simplifie                                  */
/*                   (si pas encore initialis, OA est pris pour rfrence) */
/*                  1 pour modele exact                                      */
/*  Sorties :                                                                */
/*      Qi[9]     : Longueurs calculees                                      */
/*      A[6*9]    : matrice statique =-P' dans Ro                            */
/*---------------------------------------------------------------------------*/
void Geodir(double *Qi, double *A, double *OA, int FlagExact)
{
double ABi_o[3], OBi[3], HiBi[3], FiBi[3], AiBi[3];
double h, f2, f, e, l2, l, rho2, rho;
double rctet, rstet, tetacor, Q0;
int i, i3, i6;
double *Com = OA+3;

	if(FlagExact !=0) /* Modele geometrique exact */
	{
											/* Io = [1;0;0];                */
		for (i=0; i< 9; i++)
		{
			i3 = 3*i; i6 = 2*i3;
			m3v3(ABi_o, Com, ABi+i3);		/* ABi_o = Com*ABi(:,i); 		*/
			addv3(OBi, OA, ABi_o);			/* OBi = OA + ABi_o; 			*/
			subv3(HiBi, OBi, OHi+i3);		/* HiBi = OBi - OHi(:,i); 		*/
			h = *HiBi;						/* h = HiBi'*Io; 				*/
			cpyv3(FiBi, HiBi); *FiBi -= h;	/* FiBi = HiBi-h*Io; 			*/
			f2 = carrev3(FiBi); f = sqrt(f2);/* f2 = FiBi'*FiBi ; 			*/
			/* scav3(Ji,1./f,FiBi);	*/		/* Ji = FiBi/f; 				*/
			e = f - Rpoul;
			l2 = e*e + h*h; l = sqrt(l2);
			rho2 = l2 - Rpoul*Rpoul; rho = sqrt(rho2);
			if (flagbas[i])
			{
				rctet = Rpoul*(Rpoul*h-rho*e)/l2;
				rstet = Rpoul*(Rpoul*e+rho*h)/l2;
				tetacor = atan2(-rctet,rstet);
				Q0 = Rpoul*(M_PI - tetacor);
			}
			else
			{
				rctet = Rpoul*(Rpoul*h+rho*e)/l2;
				rstet = Rpoul*(Rpoul*e-rho*h)/l2;
				tetacor = atan2(-rctet,rstet);
				Q0 = Rpoul*(M_PI + tetacor);
			}
							  /* AiBi = HiBi - (Rpoul+rstet)*Ji - rctet*Io; */
			cumv3(AiBi, HiBi, -(Rpoul+rstet)/f, FiBi); *AiBi -= rctet;
			scav3(A+i6, -1./rho, AiBi);		/* Ui = AiBi/rho;				*/
			crossv3(A+i6+3, ABi_o, A+i6);/* matA(:,i)=-[Ui;cross(ABi_o,Ui)];*/
			Qi[i] = rho + Q0;
			if (FlagExact == DEFSITINI || 1==1) /* Ini. const. modele simplifie */
			{
				Qzero[i] = Q0;
				subv3(OAi+i3, OBi, AiBi);
				FlagNotiniConstSmp = 0; /* Ini. const. modele simplifie faite */
			}
		}
/* #define PRINTOAi */
#ifdef PRINTOAi
		if (FlagExact == DEFSITINI) 
		{
			puts("Composantes des OAi :");
			for(i=0;i<9;i++)
			{
				printf(" %7.4lf %7.4lf %7.4lf\n",OAi[i*3+2], OAi[i*3+1], -OAi[i*3]);
			}
		}
#endif
	}
	else /* Modele simplifie */
	{
		if(FlagNotiniConstSmp) 
			Geodir(Qi, A, OA_ref, DEFSITINI);
		for (i=0; i< 9; i++)
		{
			i3 = 3*i; i6 = 2*i3;
			m3v3(ABi_o, Com, ABi+i3);		/* ABi_o = Com*ABi(:,i); 		*/
			addv3(OBi, OA, ABi_o);		/* OBi = OA + ABi_o; 			*/
			subv3(AiBi, OBi, OAi+i3);		/* AiBi = OBi - OAi(:,i); 		*/
			rho = longv3(AiBi);
			scav3(A+i6, -1./rho, AiBi);		/* Ui = AiBi/rho;				*/
			crossv3(A+i6+3, ABi_o, A+i6);	/* matA(:,i)=-[Ui;cross(ABi_o,Ui)];*/
			Qi[i] = rho + Qzero[i];
		}
	}
}
