#include <math.h>
#include "utils.h"
void nextcholl(double* a, double* b, int n, int q);

#define NVAR  9		                /* Programmation specifique 9 variables */
#define NEQ 6						/* Soumises  6 contraintes galits    */
#define NFREE 3                     /* Il y a donc 3 degrs de libert      */
#define MAXCAS 987	                               

int Statinv (double *Fq, double *Fxdes, double *A, double *tmin, double *tmax,
									double *tref, double *tpon, int Satur)
/*---------------------------------------------------------------------------*/
/*       Modele statique inverse : Rsolution de Fxdes = A * Fq              */
/*---------------------------------------------------------------------------*/
/*                                                                           */
/*	Sorties :																 */
/*		Fq		: Vecteur des tensions des 9 cbles							 */
/*																			 */
/*	Entres :																 */
/*		Fxdes	: Torseur cartsien  raliser								 */
/*		A		: Matrice 6 x 9  telle que  Fxdes = A * Fq 					 */
/*																			 */
/*		tmin    : Vecteur des 9 tensions minimales							 */
/*		tmax    : Vecteur des 9 tensions maximales							 */
/*		tref    : Vecteur des 9 tensions de rfrence						 */
/*		tpon    : Vecteur des 9 tensions de pondration						 */
/*                                                                           */
/*		Satur	: 0 : tmin et tmax ne sont pas pris en compte dans			 */
/*					  le calcul des Fq. Si une ou plusieurs tensions		 */
/*					  sont hors domaine, elle est sature. Fxdes ne			 */
/*					  sera pas ralis, mme s'il existe une solution		 */
/*					  qui ralise Fxdes en respectant les contraintes.		 */
/*				  1 : tmin et tmax sont pris en compte dans le calcul 		 */
/*				      des Fq. Si une solution qui ralise Fxdes en			 */
/*					  respectant les contraintes existe, elle est trouve.	 */
/*																			 */
/*    La solution minimise la somme des [Fq(i)-tref(i))/tpon(i)]^2           */
/*      tout en respectant (si Satur != 0)  tmin(i) < Fq(i) < tmax(i)        */
/*                                                                           */
/*	Retour :																 */
/*				0 : solution exacte, 0 contraintes satures					 */
/*			n > 0 : solution exacte, n contraintes satures					 */
/*			n < 0 : solution fausse, |n| contraintes satures				 */
/*			  -10 : Echec rsolution. Pas de solution.						 */
/*																			 */
/*---------------------------------------------------------------------------*/
{
int i, ii, j, jj, ij, k, jns, jNvar;
int nbcascur, nbcasnew, nsat, icasat;
int idem, flagtrouve, flagnew;
int flagVK[NVAR] = {0};
int bufindex1[MAXCAS], bufindex2[MAXCAS];
int pixcur[NFREE], *pixsat, *pixsatcur, *pixsatnew, *pixcascur, *pixcasnew;
double VK[NVAR*NVAR], tlim[NVAR], MM[NFREE*NFREE], mu[NFREE];
double AD1[NEQ*NVAR], dF[NVAR], Mat[NEQ*NEQ], Fqref[NVAR];
double delta, temp ;
int choll(double* a, double* b, int n, int q) ;
	if (!Satur)  /* On ne prend pas les contraintes en compte */
	{
		matmat(dF,A,tref,NEQ,NVAR,1); 
		subvec(dF,Fxdes,dF,NEQ);
		for (i = 0; i<NVAR; i++) scavec(AD1+i*NEQ,tpon[i],A+i*NEQ, NEQ);
		mattmat(Mat,AD1,AD1,NEQ,NVAR,NEQ);
		if (choll(dF,Mat,NEQ,1) < 0) return -10;
		tmatmat(Fq, AD1,dF,NVAR,NEQ,1);
		mulvec(Fq,tpon,Fq,NVAR);
		addvec(Fq,tref,Fq,NVAR);
		for (nsat = 0, i = 0; i<NVAR; i++)
		{
			if (Fq[i] < tmin[i]) { nsat--; Fq[i] = tmin[i];}
			else if (Fq[i]>tmax[i]) { nsat--; Fq[i] = tmax[i];}
		}
		return nsat;
	}

	/* Solution avec prise en compte des contraintes. Algorithme perso de
	        programmation quadratique adapt de la mthode de Theil et Van de Panne */	   
	
	pixsatcur = bufindex1;/* Les diffrents cas de cont. satures en cours d'examen */
	pixsatnew = bufindex2;   /* Les nouveaux cas de contraintes satures a examiner */
	nbcascur = 1;   /* Au dbut c'est le cas sans contrainte qui est le cas courant */
	nbcasnew = 0;  /* Nouveaux cas d'ensembles de contraintes simultanes  saturer */
	flagtrouve = 0;                                /* On n'a pas trouv la solution */
	nsat = 0;	                     /* Nombre decontraintes simultanment satures */

/* On traite les ensembles de contraintes satures nsat par nsat (0 par 0, 1 par 1, ...) */
	for (nsat = 0; nsat <= NFREE; nsat++) 
	{
		/* Examen de toutes les cas de contraintes satures */
		for (icasat = 0, pixcascur = pixsatcur, pixcasnew = pixsatnew; 
			 icasat < nbcascur; icasat++,   pixcascur += nsat)
		{
			/* Calcul de la solution */			
			if (nsat == 0) 
			{ /* Solution nominale sans contraintes ingalit */;					
				matmat(dF,A,tref,NEQ,NVAR,1); 
				subvec(dF,Fxdes,dF,NEQ);
				for (i = 0; i<NVAR; i++) scavec(AD1+i*NEQ,tpon[i],A+i*NEQ, NEQ);
				mattmat(Mat,AD1,AD1,NEQ,NVAR,NEQ);
				if (choll(dF,Mat,NEQ,1) < 0) return -10 ;
				tmatmat(Fq, AD1,dF,NVAR,NEQ,1);
				mulvec(Fq,tpon,Fq,NVAR);
				addvec(Fq,tref,Fq,NVAR);
				cpyvec(Fqref,Fq,NVAR);
			}
			else /* Calcul de la solution gnrale nsat + NEQ <= NVAR */
			{
				for (jj = jns = 0; jj < nsat; jj++, jns += nsat)
				{
					j =  pixcascur[jj];
					if (j < NVAR) tlim[j] = tmin[j];
					else {j -= NVAR; tlim[j] = tmax[j];}
					jNvar = j*NVAR;
					pixcur[jj] = j;
					/* Calcul des Vk si pas deja calculs */
					if (flagVK[j]==0) 
					{
						cpyvec(dF,AD1+j*NEQ,NEQ);
						nextcholl(dF,Mat,NEQ,1);
						tmatmat(VK+jNvar, AD1,dF,NVAR,NEQ,1);
						flagVK[j] = 1;
					}
					/* Constitution matrice facteur de mu */
					for (ii = 0 ; ii <= jj; ii++) /* Seule la partie suprieure est ncessaire */
					{
						i = pixcur[ii];
						ij = ii+jns;
						MM[ij] = -VK[i+jNvar];
						if (ii==jj) MM[ij] += 1.;
					}
					mu[jj] = (Fqref[j] - tlim[j])/tpon[j];
				}
				if (nsat==1) 
				{
					if (MM[0]==0.) break; /* On laisse tomber ce cas */
					mu[0] /= MM[0];
				}
				else 
				{
					if (nsat == 2)
					{
						delta = MM[0]*MM[3]-MM[2]*MM[2];
						if (delta == 0.) break ; /* On laisse tomber ce cas */
						temp  = ( MM[3]*mu[0]-MM[2]*mu[1])/delta;
						mu[1] = (-MM[2]*mu[0]+MM[0]*mu[1])/delta;
						mu[0] = temp;
					}
					else
					{
						if (choll(mu,MM,nsat,1) < 0) break ; /* On laisse tomber ce cas */
					}
				}
				cpyvec(Fq, Fqref, NVAR);
				for (ii = 0 ; ii < nsat; ii++)
				{
					i = pixcur[ii];
					mulvec(dF,tpon,VK+i*NVAR,NVAR);
					cumvec(Fq, Fq, mu[ii],dF,NVAR);
				}
				for (ii = 0 ;ii < nsat;ii++)
				{
					i = pixcur[ii];
					Fq[i] = tlim[i];
				}
			}
			flagtrouve = 1;
			/* Test si solution admissible */
			for (i=0; i < NVAR ; i++)
			{
				if (Fq[i] < tmin[i]) ii = i;
				else if (Fq[i] > tmax[i]) ii = i + NVAR;
				else continue;
				flagtrouve = 0;
				if (nsat == NFREE) {nbcasnew++; break; } /* Inutile d'en faire plus. Pb trop contraint */
				/* Constitution d'un nouveau cas de contrainte  saturer */
				for (j = 0, k= 0 ; k < nsat; k++) /* rang de la nlle contrainte / au cas en cours */
				{
					if (ii < pixcascur[k]) break; else j++; 
				}
				/* On range le nouveau cas */
				for (k=0; k < j; k++) pixcasnew[k] = pixcascur[k];
				pixcasnew[j] = ii;
				for (k = j+1; k<= nsat; k++) pixcasnew[k] = pixcascur[k-1];

				/* Test si le cas est vraiment nouveau */
				flagnew = 1 ;
				for (j = 0, pixsat = pixsatnew ; j < nbcasnew; j++, pixsat += nsat+1)
				{
					idem = 1;
					for (k = 0 ; k <= nsat ; k++) 
					{
						if(pixsat[k] != pixcasnew[k]) {idem = 0; break;}
					}
					if (idem) {flagnew = 0; break;}						
				}
				/* C'est un nouveau cas */
				if (flagnew)
				{
					nbcasnew++; 
					pixcasnew += nsat+1;
				}					
			}

			/* Test si solution optimale pour nsat > 1 (pour 0 et 1 c'est forcment optimal) */
			/* Plac aprs la conformit, car ce test est quasiment toujours vrifi, il sera
			   donc moins souvent appel ici, dans la mesure ou le prcdent est moins
			   frquemment vrifi */
			if (flagtrouve && nsat > 1) for(k=0; k < nsat; k++)
			{
				if (pixcascur[k] < NVAR) { if (mu[k] > 0.) { flagtrouve = 0; break; }}
				else { if (mu[k] < 0.) { flagtrouve = 0; break; }}
			}

			if (flagtrouve) break;
		} /* Fin boucle sur icasat */
		if (flagtrouve) break;
		/* On permute les buffeurs d'index */
		pixsat = pixsatcur;
		pixsatcur = pixsatnew;
		pixsatnew = pixsat;
		nbcascur = nbcasnew;
		nbcasnew = 0;
	}
    if (flagtrouve) return nsat ;

	/* Rien trouv. On sature la solution initiale */
	cpyvec(Fq,Fqref,NVAR);
	for (nsat = 0, i = 0; i<NVAR; i++)
	{
		if (Fq[i] < tmin[i]) { nsat--; Fq[i]=tmin[i];}
		else if (Fq[i]>tmax[i]) { nsat--; Fq[i]=tmax[i];}
	}
	return nsat ;
}	

