#include <math.h>
#include <float.h>
#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif
/*---------------------------------------------------------------------
	Recopie d'un vecteur de R3 : U = V
	----------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double *v	: Adresse vecteur donn

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect
----------------------------------------------------------------------*/
double *cpyv3(double* u, double* v)
{
    u[0] = v[0] ;
    u[1] = v[1] ;
    u[2] = v[2] ;
	return u;
}
/*---------------------------------------------------------------------
	Ngation d'un vecteurs de R3 : U = -V
	-------------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double *v	: Adresse vecteur donn

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect

    Remarque : u et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *negv3(double* u, double* v)
{
    u[0] = -v[0] ;
    u[1] = -v[1] ;
    u[2] = -v[2] ;
	return u;
}
/*---------------------------------------------------------------------
	Addition de vecteurs de R3 : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *addv3(double* u, double* v, double* w)
{
    u[0] = v[0] + w[0] ;
    u[1] = v[1] + w[1] ;
    u[2] = v[2] + w[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Soustraction de vecteurs de R3 : U = V - W
	------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *subv3(double* u, double* v, double* w)
{
    u[0] = v[0] - w[0] ;
    u[1] = v[1] - w[1] ;
    u[2] = v[2] - w[2] ;
	return u;
}
/*---------------------------------------------------------------------
	Carr d'un vecteurs de R3 : s = A'*A
	------------------------------------

	Entrees:
		double *A	: Adresse vecteur donn

	Sorties:
		Retour		: Somme des carrs des composantes
----------------------------------------------------------------------*/
double carrev3(double* a)
{
    return(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]);
}
/*---------------------------------------------------------------------
	Produit scalaire de vecteurs de R3 : s = <A . B>
	------------------------------------------------

	Entrees:
		double *A	: Adresse du premier vecteur de R3 donn
		double *B	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		Retour		: produit scalaire
----------------------------------------------------------------------*/
double dotv3(double* a, double* b)
{
    return(a[0]*b[0]+a[1]*b[1]+a[2]*b[2]);
}
/*---------------------------------------------------------------------
	Longueur d'un vecteurs de R3 : s = sqrt(A'*A)
	---------------------------------------------

	Entrees:
		double *A	: Adresse vecteur donn

	Sorties:
		Retour		: Longueur du vecteur
----------------------------------------------------------------------*/
double longv3(double* a)
{
    return(sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]));
}
/*---------------------------------------------------------------------
	Normalisation d'un vecteur de R3 : A = B / ||B||
	------------------------------------------------

	Entrees:
		double *A	: Adresse vecteur unitaire rsultat
		double *B	: Adresse vecteur donn

	Sorties:
		A est affect
		Retour		: Longueur du vecteur B

	Remarque : A et B peuvent occuper la mme mmoire
		       si ||B|| = 0  -> A[0] = 1, les autres nuls.

	Remarque : A peut occuper la mme mmoire que B
----------------------------------------------------------------------*/
double lunitv3(double* a, double* b)
{
    double l ;
    l = sqrt(b[0]*b[0]+b[1]*b[1]+b[2]*b[2]);
    if (l == 0.)
    {
        a[0]=1.; a[1]=0.; a[2]=0.;
    }
    else
    {
        a[0] = b[0]/l;
		a[1] = b[1]/l;
        a[2] = b[2]/l;
    }
    return(l);
}
/*---------------------------------------------------------------------
	Multiplication d'un vecteur de R3 par un scalaire : U = sV
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du vecteur de R3 donn
		double s	: Scalaire donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *scav3(double* u, double s, double* v)
{
    u[0] = s * v[0] ;
    u[1] = s * v[1] ;
    u[2] = s * v[2] ;
	return u;
}
/*---------------------------------------------------------------------
	Produit vectoriel de vecteurs de R3 : A = B ^ C
	-----------------------------------------------

	Entrees:
		double *A	: Adresse du vecteur affect
		double *B	: Adresse du premier vecteur de R3 donn
		double *C	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		A est affect
		Retour		: Adresse du vecteur affect

	Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *crossv3(double* a, double* b, double* c)
{
double p, q ;
      p  = b[1] * c[2] - b[2] * c[1];
      q  = b[2] * c[0] - b[0] * c[2];
    a[2] = b[0] * c[1] - b[1] * c[0];
    a[0] = p ;
    a[1] = q ;
	return a;
}
/*---------------------------------------------------------------------
	Cumulation de vecteurs de R3 : U = V + aW
	-----------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double a	: Scalaire donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *cumv3(double* u, double* v,double a, double* w)
{
    u[0] = v[0] + a*w[0] ;
    u[1] = v[1] + a*w[1] ;
    u[2] = v[2] + a*w[2] ;
	return u;
}
/*---------------------------------------------------------------------
	Recopie d'une matrice de R3 : A = B
	-----------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la matrice m3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte
----------------------------------------------------------------------*/
double *cpym3(double* A, double* B)
{
int i ; for(i=0;i<9;i++) A[i]=B[i];
	return A;
}
/*---------------------------------------------------------------------
	Transpose d'une matrice de R3 : A = A'
	--------------------------------------

	Entrees:
		double *A	: Adresse de la matrice

	Sorties:
		A est transpose
		Retour		: Adresse de A
----------------------------------------------------------------------*/
double *tm3(double *a)
{
double s ;
	s = a[3]; a[3] = a[1]; a[1] = s;
	s = a[6]; a[6] = a[2]; a[2] = s;
	s = a[7]; a[7] = a[5]; a[5] = s;
	return a ;
}
/*---------------------------------------------------------------------
    Multiplication d'une matrice de R3 par un scalaire : A = sB
	-----------------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice m3 affecte
		double *B	: Adresse de la matrice m3 donne
		double s	: Scalaire donn

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice m3 affecte

    Remarque : A et B peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *scam3(double* A, double s, double* B)
{
int i ; for(i=0;i<9;i++) A[i]=s*B[i];
	return A;
}
/*---------------------------------------------------------------------
	Produit matrice x vecteurs de R3 : w = A v
	-----------------------------------------------

	Entrees:
		double *w	: Adresse du vecteur affect
		double *A	: Adresse de la matrice de R3 donn
		double *v	: Adresse du vecteur de R3 donn

	Sorties:
		w est affect
		Retour		: Adresse du vecteur affect

    Remarque : w et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *m3v3(double* w, double* a, double* v)
{
double p, q ;
      p  = a[0] * v[0] + a[3] * v[1] + a[6] * v[2];
      q  = a[1] * v[0] + a[4] * v[1] + a[7] * v[2];
    w[2] = a[2] * v[0] + a[5] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
	return w;
}
/*---------------------------------------------------------------------
	Produit matrice transpose x vecteurs de R3 : w = A' v
	------------------------------------------------------

	Entrees:
		double *w	: Adresse du vecteur affect
		double *A	: Adresse de la matrice de R3 donn
		double *v	: Adresse du vecteur de R3 donn

	Sorties:
		w est affect
		Retour		: Adresse du vecteur affect

    Remarque : w et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *tm3v3(double* w, double* a, double* v)
{
double p, q ;
      p  = a[0] * v[0] + a[1] * v[1] + a[2] * v[2];
      q  = a[3] * v[0] + a[4] * v[1] + a[5] * v[2];
    w[2] = a[6] * v[0] + a[7] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
	return w;
}
/*---------------------------------------------------------------------
	Produit matrice x matrice de R3 : A = B C
	-----------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la premire matrice de R3 donne
		double *C	: Adresse de la deuxime matrice de R3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *m3m3(double* a,double* b, double* c)
{
register int i ;
double d[8] ;
    d[0] = b[0] * c[0] + b[3] * c[1] + b[6] * c[2];
    d[1] = b[1] * c[0] + b[4] * c[1] + b[7] * c[2];
    d[2] = b[2] * c[0] + b[5] * c[1] + b[8] * c[2];

    d[3] = b[0] * c[3] + b[3] * c[4] + b[6] * c[5];
    d[4] = b[1] * c[3] + b[4] * c[4] + b[7] * c[5];
    d[5] = b[2] * c[3] + b[5] * c[4] + b[8] * c[5];

    d[6] = b[0] * c[6] + b[3] * c[7] + b[6] * c[8];
    d[7] = b[1] * c[6] + b[4] * c[7] + b[7] * c[8];
    a[8] = b[2] * c[6] + b[5] * c[7] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
	return a;
}
/*---------------------------------------------------------------------
	Inverse d'une matrice de R3 : inv(A)
	------------------------------------

	Entrees:
		double *A	: Adresse matrice

	Sorties:
	    A = inv(A) si det != 0.
	Retour			: Adresse de A

   ATTENTION : EXTREMEMENT DANGEREUX SI DETERMINANT NUL
   FAIRE PRECEDER DU CALCUL DU DETERMINANT
----------------------------------------------------------------------*/
double *invm3(double* a)
{
double delta ;
double c[9] ;

    crossv3 (c, a+3, a+6) ;
    delta = dotv3(a, c) ;
    if (delta == 0.) return a ;
    crossv3 (c+3, a+6, a) ;
    crossv3 (c+6, a, a+3) ;
    tm3(scam3 (a, 1./delta, c)) ;
    return a ;
}
/*---------------------------------------------------------------------
	Gnration matrice de rotation d'angle a autour du vecteur V
	------------------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affecte
		double a	: Angle de rotation  (radians)
		double *v	: Adresse du vecteur unitaire port
		              par l'axe de rotation

	Sorties:
		m est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *av3tomrot(double* m, double a, double* v)
{
    double wx, wy, wz;
    double k;

    wx = v[0];
    wy = v[1];
    wz = v[2];
    k = 1.-cos(a);
    m[0] = 1. + k * (wx * wx - 1.);
    m[4] = 1. + k * (wy * wy - 1.);
    m[8] = 1. + k * (wz * wz - 1.);
    m[1] = k * wx * wy;
    m[2] = k * wz * wx;
    m[5] = k * wy * wz;
    k = sin(a);
    wx = k * wx;
    wy = k * wy;
    wz = k * wz;
    m[7] = m[5] - wx;
    m[5] += wx;
    m[6] = m[2] + wy;
    m[2] -= wy;
    m[3] = m[1] - wz;
    m[1] += wz;
	return m;
}
/*---------------------------------------------------------------------
	Gnration matrice de rotation autour du vecteur , d'angle gal
	                          la longeur de V
	---------------------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affecte
		double *v	: Adresse du vecteur rotation

	Sorties:
		m est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *v3tomrot(double* m, double* v)
{
double a, u[3];

	a = lunitv3(u,v) ;
	return av3tomrot(m, a, u);
}
/*---------------------------------------------------------------------
	Gnration tenseur d'inertie -mu*(V^)**2
	-------------------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double m    : masse
		double *V	: Adresse vecteur donn

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *mMuV2(double *a, double m, double *v)
{
double w[3], v02, v12, v22;

	w[0] = m*v[0]; 
	w[1] = m*v[1]; 
	w[2] = m*v[2]; 

    v02 = w[0]*v[0]; 
    v12 = w[1]*v[1]; 
    v22 = w[2]*v[2];

    a[0] = v22+v12 ; a[3] = -w[1]*v[2]; a[6] = -w[0]*v[2];
    a[1] = a[3];     a[4] =  v22+v02;   a[7] = -w[1]*v[2];
    a[2] = a[6];     a[5] =  a[7];      a[8] =  v12+v02;

	return a ;
}
/*---------------------------------------------------------------------
	Recopie d'un vecteur de Rn : U = V
	----------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double *v	: Adresse vecteur donn
		int n		: dimension de v

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect
----------------------------------------------------------------------*/
double *cpyvec (double* u, double* v, int n)
{
    if (n > 0) while (n--) u[n] = v[n];
	return u;
}
/*---------------------------------------------------------------------
	Addition de vecteurs de Rn : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *addvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] + w[n];
	return u;
}
/*---------------------------------------------------------------------
	Soustraction de vecteurs de Rn : U = V - W
	------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *subvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] - w[n];
	return u;
}
/*---------------------------------------------------------------------
	Norme infini d'un vecteurs de Rn : s = max(|Ui|)
	------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: Composante maximale du vecteur, en valeur absolue
----------------------------------------------------------------------*/
double ninfvec (double* u, int n)
{
double s = -DBL_MAX;
    if (n > 0) while (n--) if (fabs(u[n]) > s) s = fabs(u[n]);
	return s;
}
/*---------------------------------------------------------------------
	Multiplication d'un vecteur de Rn par un scalaire : U = aV
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double a	: Scalaire donn
		double *V	: Adresse du vecteur de Rn donn
		int n		: dimension du vecteur

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *scavec (double *u, double a, double *v, int n)
{
    if (n > 0) while (n--) u[n] = a * v[n] ;
	return u;
}
/*---------------------------------------------------------------------
	Addition de vecteurs de Rn : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double s	: Scalaire donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *cumvec (double* u, double* v, double s, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] + s*w[n];
	return u;
}

/*---------------------------------------------------------------------
	Multiplication terme  terme de vecteurs de Rn : U = V * W
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *mulvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] * w[n];
	return u;
}
/*---------------------------------------------------------------------
	Produit de matrice : Amp = Bmn x Cnp
	------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m	: Nombre de lignes de A et de B
		int    n	: Nombre de colonnes de B et de lignes de C
		int    p	: Nombre de colonnes de A et de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *matmat (double *A, double *B, double *C, int m, int n, int p)
{
double *a, *b, *c, r ;
int i, j, k ;


    for (i = 0 ; i < m ; ++i)
    {
	a = A + i ;      			/* A(i,0) */
	c = C ;                                 /* C(0,0) */
	for (j = 0 ; j < p ; ++j)
	{
	    r = 0 ;
	    b = B + i ;				/* B(i,0) */
	    for (k = 0 ; k < n ; ++k)
	    {
		r += *b * *c++ ;		/* B(i,k) * C(k,j) */
		b += m ;
	    }
	    *a = r ;
	    a += m ;				/* A(i,j) */
	}
    }
	return A;
}
/*---------------------------------------------------------------------
	Produit de matrice : Amp = [B']mn x Cnp
	---------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m         : Nombre de lignes de A et de colonnes de B
		int    n         : Nombre de lignes de B et  de C
		int    p         : Nombre de colonnes de A et de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *tmatmat (double *a, double *b, double *c, int m, int n, int p)
{
int i, j, k, ij, ki, kj, kj0 ;

    for (j = 0, ij = 0, kj0 = 0 ; j < p ; j++, kj0 += n)
    {
	for (i = 0, ki = 0 ; i < m ; i++, ij++)
	{
	    a[ij] = 0. ;
	    for (k = 0, kj = kj0 ; k < n ; k++) a[ij] += b[ki++] * c[kj++] ;
	}
    }
	return a;
}
/*---------------------------------------------------------------------
	Produit de matrice : Amp = Bmn x [C']np
	---------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m         : Nombre de lignes de A et de lignes de B
		int    n         : Nombre de colonnes de B et de colonnes de C
		int    p         : Nombre de colonnes de A et de lignes de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *mattmat (double *a, double *b, double *c, int m, int n, int p)
{
int i, j, k, ij, ik, jk ;

    ij = -1 ;
    for (j = 0 ; j < p ; j++)
	for (i = 0 ; i < m ; i++)
	{
	    a[++ij] = 0. ;
		for (k = 0, ik = i, jk = j ; k < n ; k++, ik += m, jk += p)
		    a[ij] += b[ik]*c[jk] ;
	}
	return a;
}
/*---------------------------------------------------------------------
	Gnration matrice de rotation d'angles aronautique
				   Cap (Oz) , Assiette (Oy1) Roulis (Ox2)
	-----------------------------------------------------

	Entrees:
		double *a	: Adresse matrice affecte
		double *carol	: Adresse des angles C,A,R (radians)

	Sorties:
		a est calcule
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *carol2mrot(double *a, double *carol)
{
double cc, sc, ca, sa, cr, sr ;
double sasr, sacr ;

    cc = cos (carol[0]) ; ca = cos (carol[1]) ; cr = cos (carol[2]) ;
    sc = sin (carol[0]) ; sa = sin (carol[1]) ; sr = sin (carol[2]) ;
    sasr = sa * sr ; sacr = sa * cr ;

    a[0] = cc*ca  ; a[3] = cc*sasr - sc*cr ; a[6] = cc*sacr + sc*sr ;
    a[1] = sc*ca  ; a[4] = sc*sasr + cc*cr ; a[7] = sc*sacr - cc*sr ;
    a[2] =   -sa  ; a[5] =    ca*sr        ; a[8] =    ca*cr        ;
	return a;
}

/***********************************************************************
c ROUTINE CHOLL :  RESOLUTION SYSTEME LINEAIRE SYMETRIQUE PAR CHOLESKI
c
c                       index = choll(a,b,n,q)
c
c                       -1
c              [C] = [B]  *  [A]  avec [B] matrice symetrique
c
c    Remarque: Seule la partie superieure de B est utilisee.
c
c    Le resultat [C] est renvoye dans A
c
c    Entree :  n        Ordre du systeme
c              q        Nombre de seconds membres
c              a(n,q)   Valeurs des seconds membres
c              b(n,n)   Matrice symetrique du systeme
c
c    Sortie:   a(n,q)   Solution du systeme.
c              index    indicateur de validite du resultat.
c                                0 resultat ok.
c                               -1 pas de solution (diag=0)
c                                >0 : nb d'inversions de signe.
c
c    Methode : Algorithme de Cholevski lineaire sans racine carree:
c
c         L' equation  A = B . C   avec B symetrique peut s'ecrire :
c
c                      A = R'.D.R. C  avec:
c
c             R matrice triangulaire superieure a diagonale unite
c             D matrice diagonale.
c
c         La resolution du probleme
c                      A = R'.D.R.C
c         est alors ramenee a la resolution de
c                      A = R'. E
c         puis
c                      (E/D) = R . C
c
c         Les elements de la partie superieure de la matrice R sont calcules
c         et ranges dans la partie superieure de B.
c         Les elements de D sont calcules simultanement et ranges dans la
c         la diagonale de B.
c         Les elements de E sont calcules et ranges dans A.
c         Les elements de C sont calcules et ranges dans A.
c
c
*************************************************************************/
int choll(double* a, double* b, int n, int q)
{
int i, j, k, l, invs;
double diag, term;
double *pa, *pb, *p1, *p2, *p3;

        invs = 0;
/*
Calcul des elements Rij et Dii en lieu et place de B
*/
        pb = b;                         /* b(i,i) pour i=1,... */
        for (i = 0; i < n; ++i)
        {
           diag = *pb;                  /* diag = b(i,i) */
           if (i)
           {
              p1 = b;                   /* b(k,k) pour k=1,... */
	      p2 = pb -i;               /* b(k,i) pour k=1,... */
              for (k = 0; k < i; ++k)
	      { /* diag = diag - b(k,k)*b(k,i)*b(k,i) */
		diag -= *p1 * *p2 * *p2;
                p1 += n + 1;
                ++p2;
              }
	   }
           if (diag == 0.) return(-1);
           if (diag < 0.) ++invs;
	   *pb = diag;

           p3 = pb + n;                 /* b(i,j) pour j=i+1,... */
           for (j = i+1; j < n; ++j)
           {
              term = *p3;               /* term = b(i,j) */
              if (i)
              {
                 p1 = b;                /* b(k,k) pour k=1,... */
                 p2 = pb - i;           /* b(k,i) pour k=1,... */
		 p3 -= i;                       /* b(k,j) pour k=1,... */
                 for (k = 0; k < i; ++k)
		 {
		 /* term = term - b(k,i)*b(k,k)*b(k,j) */
                    term -= *p2++ * *p1 * *p3++;
                    p1 += n + 1;
                 }
	      }
              *p3 = term/diag;          /* b(i,j) = term / diag */
              p3 += n;
	   }
           pb += n + 1;                 /* b(i+1,i+1) */
        }
/*
calcul des elements de E en lieu et place de A
*/
        pa = a;                         /* a(1,l) pour l=1,... */
        for (l = 0; l < q; ++l)
        {
           p1 = pa;                     /* a(i,l) pour i=1,... */
	   pb = b;                      /* b(1,i) pour i=1,... */
           for (i = 0; i < n; ++i)
	   {
	      term = *p1;               /* term = a(i,l) */
              if(i)
              {
                p2 = pb;                /* b(k,i) pour k=1,... */
		p3 = pa;                /* a(k,l) pour k=1,... */
                for (k = 0; k < i; ++k)
                   /* term = term - b(k,i)*a(k,l) */
		   term -= *p2++ * *p3++;
              }
              *p1++ = term;             /* a(i,l) = term, puis a(i+1,l)*/
              pb += n;                  /* b(1,i+1) */
           }
/*
calcul des elements de C en lieu et place de A
*/
           p1 = pa + n - 1;             /* a(i,l) pour i=n,n-1,... */
           pb = p2;                     /* b(i,i) pour i=n,n-1,... */
	   for (i = n; i > 0; --i)
           {
	      term = *p1 / *pb;         /* term = a(i,l)/b(i,i) */
	      if(i < n)
              {
                p2 = pb + n;            /* b(i,i+1) */
                p3 = p1 + 1;            /* a(i+1,l) */
		for (k = i+1; k <= n; k++)
                {/* term = term - b(i,k)*a(k,l) */
                   term -= *p2 * *p3++;
		   p2 += n;             /* b(i,k+1) */
                }
              }
              *p1-- = term;             /* a(i,l) = term, puis a(i-1,l)*/
              pb -= n + 1;              /* b(i-1,i-1) */
           }
           pa += n;                     /* a(1,l+1) */
        }
	return(invs);
}

/***********************************************************************
  ROUTINE debcholl :  Prparation de la rsolution par cholesky
c
c                       index = choll(b,n)
c
c    Remarque: Seule la partie superieure de B est utilisee.
c
c    Entree :  n        Ordre du systeme
c              b(n,n)   Matrice symetrique du systeme
c
c    Sortie:   b(n,n)   Solution du systeme.
c              index    indicateur de validite du resultat.
c                                0 resultat ok.
c                               -1 pas de solution (diag=0)
c                                >0 : nb d'inversions de signe.
c
c    Methode : Algorithme de Cholevski lineaire sans racine carree:
c
c         B symetrique est dcompose en R'.D.R :
c
c             R matrice triangulaire superieure a diagonale unite
c             D matrice diagonale.
c
c         Les elements de la partie superieure de la matrice R sont calcules
c         et ranges dans la partie superieure de B.
c         Les elements de D sont calcules simultanement et ranges dans la
c         la diagonale de B.
c
c
*************************************************************************/
int debcholl(double* b, int n)
{
int i, j, k, invs;
double diag, term;
double *pb, *p1, *p2, *p3;

        invs = 0;
/*
Calcul des elements Rij et Dii en lieu et place de B
*/
        pb = b;                         /* b(i,i) pour i=1,... */
        for (i = 0; i < n; ++i)
        {
           diag = *pb;                  /* diag = b(i,i) */
           if (i)
           {
              p1 = b;                   /* b(k,k) pour k=1,... */
	      p2 = pb -i;               /* b(k,i) pour k=1,... */
              for (k = 0; k < i; ++k)
	      { /* diag = diag - b(k,k)*b(k,i)*b(k,i) */
		diag -= *p1 * *p2 * *p2;
                p1 += n + 1;
                ++p2;
              }
	   }
           if (diag == 0.) return(-1);
           if (diag < 0.) ++invs;
	   *pb = diag;

           p3 = pb + n;                 /* b(i,j) pour j=i+1,... */
           for (j = i+1; j < n; ++j)
           {
              term = *p3;               /* term = b(i,j) */
              if (i)
              {
                 p1 = b;                /* b(k,k) pour k=1,... */
                 p2 = pb - i;           /* b(k,i) pour k=1,... */
		 p3 -= i;                       /* b(k,j) pour k=1,... */
                 for (k = 0; k < i; ++k)
		 {
		 /* term = term - b(k,i)*b(k,k)*b(k,j) */
                    term -= *p2++ * *p1 * *p3++;
                    p1 += n + 1;
                 }
	      }
              *p3 = term/diag;          /* b(i,j) = term / diag */
              p3 += n;
	   }
           pb += n + 1;                 /* b(i+1,i+1) */
        }

	return(invs);
}

/**********************************************************************
c ROUTINE NextCHOLL:RESOLUTION SYSTEME LINEAIRE SYMETRIQUE PAR CHOLESKI
c
c                       index = choll(a,b,n,q)
c
c   Rsolution ultrieure de [B]*[C] = [A] qui suit un premier appel
c   de la routine CHOLL.
C   A cet appel on suppose que b contient les lments de D et R calculs
c   lors de l'appel prcdent.
c   Le resultat [C] est renvoye dans A. [B] est inchang.
c
c    Entree :  n        Ordre du systeme
c              q        Nombre de seconds membres
c              a(n,q)   Valeurs des seconds membres
c              b(n,n)   lments de D et R calculs prcdemment calculs
c
c    Sortie:   a(n,q)   Solution du systeme.
c              index    indicateur de validite du resultat.
c                                0 resultat ok.
c                               -1 pas de solution (diag=0)
c                                >0 : nb d'inversions de signe.
c
c    Methode : Algorithme de Cholevski lineaire sans racine carree:
c
c         L' equation  A = B . C   avec B symetrique peut s'ecrire :
c
c                      A = R'.D.R. C  avec:
c
c             R matrice triangulaire superieure a diagonale unite
c             D matrice diagonale.
c
c         La resolution du probleme
c                      A = R'.D.R.C
c         est alors ramenee a la resolution de
c                      A = R'. E
c         puis
c                      (E/D) = R . C
c
c         Les elements de la partie superieure de B contiennent la matrice R
c         Les elements de D dans la diagonale de B.
c         Les elements de E sont calcules et ranges dans A.
c         Les elements de C sont calcules et ranges dans A.
c
c
*************************************************************************/
void nextcholl(double* a, double* b, int n, int q)
{
int i, k, l;
double term;
double *pa, *pb, *p1, *p2, *p3;

/*
calcul des elements de E en lieu et place de A
*/
        pa = a;                         /* a(1,l) pour l=1,... */
        for (l = 0; l < q; ++l)
        {
           p1 = pa;                     /* a(i,l) pour i=1,... */
	   pb = b;                      /* b(1,i) pour i=1,... */
           for (i = 0; i < n; ++i)
	   {
	      term = *p1;               /* term = a(i,l) */
              if(i)
              {
                p2 = pb;                /* b(k,i) pour k=1,... */
		p3 = pa;                /* a(k,l) pour k=1,... */
                for (k = 0; k < i; ++k)
                   /* term = term - b(k,i)*a(k,l) */
		   term -= *p2++ * *p3++;
              }
              *p1++ = term;             /* a(i,l) = term, puis a(i+1,l)*/
              pb += n;                  /* b(1,i+1) */
           }
/*
calcul des elements de C en lieu et place de A
*/
           p1 = pa + n - 1;             /* a(i,l) pour i=n,n-1,... */
           pb = p2;                     /* b(i,i) pour i=n,n-1,... */
	   for (i = n; i > 0; --i)
           {
	      term = *p1 / *pb;         /* term = a(i,l)/b(i,i) */
	      if(i < n)
              {
                p2 = pb + n;            /* b(i,i+1) */
                p3 = p1 + 1;            /* a(i+1,l) */
		for (k = i+1; k <= n; k++)
                {/* term = term - b(i,k)*a(k,l) */
                   term -= *p2 * *p3++;
		   p2 += n;             /* b(i,k+1) */
                }
              }
              *p1-- = term;             /* a(i,l) = term, puis a(i-1,l)*/
              pb -= n + 1;              /* b(i-1,i-1) */
           }
           pa += n;                     /* a(1,l+1) */
        }
}